#!/usr/bin/perl 
#
# mailaddrcrypt-server
# author, (c): Philippe Kueck <projects at unixadm dot org>
#

use strict;
use warnings;

use POSIX qw(setsid);
use IO::Select;
use IO::Socket;
use MIME::Base32;
use Sys::Syslog qw(:standard :macros);

use Pod::Usage;
use Getopt::Long;

my ($sock, $opts, $domains);

sub slogf {
	if ($opts->{'verbose'}) {
		printf STDERR "%s: ".$_[1]."\n", $_[0], $_[2];
		return
	}
	syslog($_[0], sprintf $_[1], $_[2])
}

sub parsecfg {
	my $cf = $opts->{'config'} || $_[1];
	return printf STDERR "cannot read file %s.\n", $cf unless -r $cf;
	undef $domains;
	my $block;
	open CFG, "< $_[1]" or return;
	while (<CFG>) {
		s/(?:^\s+|\s+$)//g; s/^#.*//; next unless $_;
		if (!$block && /^@?([\w.-]+)\s*{/) {$block = $1; next}
		undef $block if /^}/; next unless $block;
		if ($block eq 'global') {
			if (/^(user|group)\s*=\s*(\w+)$/) {$opts->{$1} = $2}
			if (/^(port|sockmask)\s*=\s*(\d{1,5})$/) {$opts->{$1} = $2}
			if (/^(socket|pidfile)\s*=\s*(\S+)$/) {$opts->{$1} = $2}
			next
		}
		next unless /^(\S+)\s*=\s*([@\w_.-]+)$/;
		$domains->{$block}->{$1} = $2
	}
	close CFG;
	$opts->{'config'} = $cf
}

sub writepid {
	return unless $opts->{'pidfile'};
	open PID, ">". $opts->{'pidfile'} or die $@;
	print PID $_[0];
	close PID
}

sub daemonize {
	chdir '/';
	defined (my $pid = fork) or die "cannot fork $!";
	if ($pid) {
		writepid $pid;
		exit
	}
	setsid
}

sub dropprivs {
	if (defined $opts->{'group'}) {
		if (my $gid = getgrnam($opts->{'group'})) {
			$( = $) = "$gid $gid";
			die "Cannot drop group privileges" unless $> == $gid
		} else {
			die "Unknown group $opts->{'group'}"
		}
	}
	if (defined $opts->{'user'}) {
		if (my $uid = getpwnam($opts->{'user'})) {
			$> = $< = $uid;
			die "Cannot drop user privileges" unless $> == $uid
		} else {
			die "Unknown user $opts->{'user'}"
		}
	}
	umask 0
}

sub decrypt {
	my ($c, $d) = @_;
	foreach (keys %{$domains->{$d}}) {
		my $k = $_ . "\0";
		$k .= substr($k, 0, length($c) - length($k)) while length($k) < length($c);
		$c .= substr($c, 0, length($k) - length($c)) while length($c) < length($k);
		return $domains->{$d}->{$_} if ($k ^ $c) =~ /^[\x20-\x7e]+\0/
	}
	return
}

sub bailout {
	$sock->close if $sock;
	unlink $opts->{'socket'} if $opts->{'socket'};
	unless ($opts->{'verbose'}) {
		syslog("info", "terminating");
		closelog()
	}
}

Getopt::Long::Configure("no_ignore_case");
GetOptions(
	'v|verbose' => \$opts->{'verbose'},
	'd|debug' => sub {$opts->{'debug'} = $opts->{'verbose'} = 1},
	'u|user=s' => \$opts->{'user'},
	'g|group=s' => \$opts->{'group'},
	'p|port=i' => \$opts->{'port'},
	's|socket=s' => \$opts->{'socket'},
	'pid=s'=> \$opts->{'pidfile'},
	'c|config=s' => \&parsecfg,
	'f|foreground' => \$opts->{'foreground'},
	'h|help' => sub {pod2usage({'-exitval' => 0, '-verbose' => 2})}
) or pod2usage({'-exitval' => 1, '-verbose' => 0});

pod2usage({'-exitval' => 1, '-verbose' => 0}) unless keys %$domains;

unless ($opts->{'verbose'}) {
	if ($opts->{'foreground'}) {writepid $$}
	else {&daemonize}
	&dropprivs;
	$0 = "mailaddrcrypt-server";
	$SIG{'HUP'} = \&parsecfg;
	openlog("mailaddrcrypt-server", "ndelay,pid", LOG_MAIL);
	syslog("info", "starting up")
}
$SIG{'INT'} = \&bailout;

$|++;
for (;;) {
	syslog("info", "listening");
	if ($opts->{'socket'}) {
		unlink $opts->{'socket'} if -e $opts->{'socket'};
		$sock = new IO::Socket::UNIX(
			'Local' => $opts->{'socket'}, 'Type' => SOCK_DGRAM,
			'Listen' => SOMAXCONN
		);
		chmod $opts->{'sockmask'} if $opts->{'sockmask'} &&
			$opts->{'sockmask'} =~ /^0?[0-7]{3}$/
	} else {
		$sock = new IO::Socket::INET(
			'LocalHost' => '127.0.0.1',
			'LocalPort' => $opts->{'port'} || 2525,
			'Proto' => 'tcp', 'Listen' => SOMAXCONN,
			'ReuseAddr' => 1
		)
	}
	unless ($sock) {
		slogf "err", "cannot open socket %s", $@;
		exit 1
	}
	my $sel = new IO::Select($sock);

	while (my @rdy = $sel->can_read) {
		foreach my $csock (@rdy) {
			if ($csock == $sock) {$sel->add($sock->accept); next}
			$_ = <$csock>;
			unless ($_) {$sel->remove($csock); close($csock); next}
			s/[\r\n]+$//g;
			printf STDERR "<<< [%s]\n", $_ if $opts->{'debug'};
			unless (/^get\s+([a-z2-7]+)@(\S+)/i && exists $domains->{lc $2}) {
				printf STDERR ">>> 500 DUNNO\n" if $opts->{'debug'};
				printf $csock "500 DUNNO\r\n";
				next
			}
			my $acct = decrypt(MIME::Base32::decode_rfc3548(uc($1)), lc $2);
			if (defined $acct) {
				printf STDERR ">>> [200 %s]\n", $acct if $opts->{'debug'};
				printf $csock "200 %s\r\n", $acct;
				next
			}
			slogf "warning", "invalid address [%s\@%s]", $1, $2;
			printf STDERR ">>> 500 REJECT\n" if $opts->{'debug'};
			printf $csock "500 REJECT Invalid address.\r\n"
		}
	}
	$sock->close
}

&bailout;


__END__

=head1 NAME

mailaddrcrypt-server

=head1 VERSION

$Revision$

=head1 SYNOPSIS

mailaddrcrypt-server [OPTIONS]

=head1 OPTIONS

=over 8

=item B<-c>,B<--config> I<FILE>

Mandatory. Read config from I<FILE>.

=item B<-u>,B<--user> I<USER>

Drop user privileges to I<USER>.

=item B<-g>,B<--group> I<GROUP>

Drop group privileges to I<GROUP>.

=item B<-p>,B<--port> I<PORT>

Bind to port I<PORT> (defaults to 2525 if not specified).

=item B<-s>,B<--socket> I<FILE>

Bind to unix socket I<FILE>.

=item B<--pid> I<FILE>

Save pid to I<FILE>.

=item B<-f>,B<--foreground>

Don't fork to background.

=item B<-v>,B<--verbose>

Be verbose and don't fork to background.

=item B<-d>,B<--debug>

Debug communication between client and server.

=back

=head1 DESCRIPTION

This daemon expects C<get> and an email address. It validates the localpart and checks whether or not it has been generated with C<mailaddrcrypt>. If the I<KEY> matches, it returns the corresponding I<ACCOUNT>. The localpart is a base32 encoded string of the key XOR some description text.

=head1 CONFIG FILE

 options {
     user = nobody
     group = nobody
     pidfile = /var/run/mailaddrcrypt-server.pid
     port = 2525
     socket = /run/mailaddrcrypt/server.sock
     sockmask = 660
 }

 @yourdomain.tld {
     KEY1 = ACCOUNT1
     KEY2 = ACCOUNT2
 }

 @anotherdomain.tld {
     KEY3 = ACCOUNT3
 }

=head1 DEPENDENCIES

=over 8

=item B<Pod::Usage>

=item B<Getopt::Long>

=item B<IO::Select>

=item B<IO::Socket>

=item B<MIME::Base32>

=item B<Sys::Syslog>

=back

=head1 AUTHOR

Philippe Kueck <projects at unixadm dot org>

=cut
